import re
from dataclasses import dataclass, field
from typing import List, Optional

from .constants import order_to_publish

from ..adders import add_reference_order
from ..getters import get_data, get_options
from ..models import Relatant

@dataclass
class ModelProperties:
    Time: Optional[str]
    Space: Optional[str]
    
    @classmethod
    def from_query(cls, raw_data: List) -> 'ModelProperties':

        data = raw_data[0]

        Time = ''
        if data.get('isTimeContinuous', {}).get('value') == 'True':
            Time = 'continuous'
        elif data.get('isTimeDiscrete', {}).get('value') == 'True':
            Time = 'discrete'

        Space = ''
        if data.get('isSpaceContinuous', {}).get('value') == 'True':
            Space = 'continuous'
        elif data.get('isSpaceDiscrete', {}).get('value') == 'True':
            Space = 'discrete'

        return cls(
            Time = Time,
            Space = Space,
        )
    
@dataclass
class Variables:
    ID: Optional[str]
    Name: Optional[str]
    Unit: Optional[str]
    Symbol: Optional[str]
    Task: Optional[str]
    Type: Optional[str]
    
    @classmethod
    def from_query(cls, data: List) -> 'Variables':

        return cls(
            ID = data.get('ID', {}).get('value'),
            Name = data.get('Name', {}).get('value'),
            Unit = data.get('Unit', {}).get('value'),
            Symbol = re.sub(r'(<math\b(?![^>]*\bdisplay=)[^>]*)(>)', r'\1 display="inline"\2', data.get('Symbol', {}).get('value')),
            Task = data.get('label', {}).get('value'),
            Type = data.get('Type', {}).get('value')
        )
    
@dataclass
class Parameters:
    Name: Optional[str]
    Unit: Optional[str]
    Symbol: Optional[str]
    Task: Optional[str]
    
    @classmethod
    def from_query(cls, data: List) -> 'Parameters':

        return cls(
            Name = data.get('Name', {}).get('value'),
            Unit = data.get('Unit', {}).get('value'),
            Symbol = re.sub(r'(<math\b(?![^>]*\bdisplay=)[^>]*)(>)', r'\1 display="inline"\2', data.get('Symbol', {}).get('value')),
            Task = data.get('label', {}).get('value')
        )
    
@dataclass
class MRelatant:
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    
    @classmethod
    def from_query(cls, raw: str) -> 'MRelatant':

        id_mathalgodb = ''
        id_general, label, description, link = raw.split(" | ")

        if link and link.startswith('https://mardi4nfdi.de/mathalgodb/0.1/algorithm#'):
            _, id_mathalgodb = link.split('#')

        return cls(
            id = f"mathalgodb:{id_mathalgodb}" if id_mathalgodb else id_general,
            label = label,
            description = description,
        )
    
@dataclass
class ProcessStep:
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    inputDataSet: Optional[List[Relatant]] = field(default_factory=list)
    outputDataSet: Optional[List[Relatant]] = field(default_factory=list)
    uses: Optional[List[Relatant]] = field(default_factory=list)
    platformSoftware: Optional[List[Relatant]] = field(default_factory=list)
    platformInstrument: Optional[List[Relatant]] = field(default_factory=list)
    fieldOfWork: Optional[List[Relatant]] = field(default_factory=list)
    mscID: Optional[List[Relatant]] = field(default_factory=list)

    
    @classmethod
    def from_query(cls, raw_data: List) -> 'ProcessStep':

        data = raw_data[0]
        
        # Load MSC Classification
        msc = get_data('data/msc2020.json')

        return cls(
            id = None,
            label = None,
            description = None,
            inputDataSet = [Relatant.from_query(input) for input in data.get('inputDataSet', {}).get('value', '').split(" / ") if input] if 'inputDataSet' in data else [],
            outputDataSet = [Relatant.from_query(output) for output in data.get('outputDataSet', {}).get('value', '').split(" / ") if output] if 'outputDataSet' in data else [],
            uses = [MRelatant.from_query(algorithm) for algorithm in data.get('uses', {}).get('value', '').split(" / ") if algorithm] if 'uses' in data else [], 
            platformSoftware = [Relatant.from_query(software) for software in data.get('platformSoftware', {}).get('value', '').split(" / ") if software] if 'platformSoftware' in data else [],
            platformInstrument = [Relatant.from_query(instrument) for instrument in data.get('platformInstrument', {}).get('value', '').split(" / ") if instrument] if 'platformInstrument' in data else [], 
            fieldOfWork = [Relatant.from_query(field) for field in data.get('fieldOfWork', {}).get('value', '').split(" / ") if field] if 'fieldOfWork' in data else [],
            mscID = [Relatant.from_msc(mscID, next((key for key, value in msc.items() if value['id'] == mscID), None), next((value['quote'] for key, value in msc.items() if value['id'] == mscID), None)) for mscID in data.get('mscID', {}).get('value', '').split(" / ") if mscID] if 'mscID' in data else [],
        )

@dataclass
class Method:
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    implementedBySoftware: Optional[List[Relatant]] = field(default_factory=list)
    implementedByInstrument: Optional[List[Relatant]] = field(default_factory=list)

    @classmethod
    def from_query(cls, raw_data: List) -> 'ProcessStep':

        data = raw_data[0]
        
        return cls(
            id = None,
            label = None,
            description = None,
            implementedBySoftware = [Relatant.from_query(software) for software in data.get('implementedBySoftware', {}).get('value', '').split(" / ") if field] if 'implementedBySoftware' in data else [],
            implementedByInstrument = [Relatant.from_query(software) for software in data.get('implementedByInstrument', {}).get('value', '').split(" / ") if field] if 'implementedByInstrument' in data else [] 
        )
    
@dataclass
class Software:
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    sourceCodeRepository: Optional[str]
    userManualURL: Optional[str]
    reference: Optional[List] = field(default_factory=list)
    programmedIn: Optional[List[Relatant]] = field(default_factory=list)
    dependsOnSoftware: Optional[List[Relatant]] = field(default_factory=list)
    
    @classmethod
    def from_query(cls, raw_data: dict) -> 'Software':

        data = raw_data[0]
        order = add_reference_order('software')

        return cls(
            id = None,
            label = None,
            description = None,
            sourceCodeRepository = data.get('sourceCodeRepository', {}).get('value', '') ,
            userManualURL = data.get('userManualURL', {}).get('value', ''),
            reference = {order[key][0]: [order[key][1], value] for part in data.get('reference', {}).get('value', '').split(' | ') if (key := part.split(':')[0]) in order and (value := part.split(':')[1])} | ({order['url'][0]: [order['url'][1], url]} if (url := next((part for part in data.get('reference', {}).get('value', '').split(' | ') if part.startswith('https://')), None)) else {}),
            programmedIn =  [Relatant.from_query(language) for language in data.get('programmedIn', {}).get('value', '').split(" / ") if language] if 'programmedIn' in data else [], 
            dependsOnSoftware = [Relatant.from_query(software) for software in data.get('dependsOnSoftware', {}).get('value', '').split(" / ") if software] if 'dependsOnSoftware' in data else []
        )
    
@dataclass
class Hardware:
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    nodes: Optional[str]
    cores: Optional[str]
    CPU: Optional[List[Relatant]] = field(default_factory=list)
    
    @classmethod
    def from_query(cls, raw_data: dict) -> 'Hardware':

        data = raw_data[0]

        return cls(
            id = None,
            label = None,
            description = None, 
            nodes = data.get('nodes', {}).get('value', ''),
            cores = data.get('cores', {}).get('value', ''),
            CPU = [Relatant.from_query(cpu) for cpu in data.get('CPU', {}).get('value', '').split(" / ") if cpu] if 'CPU' in data else []
        )

@dataclass
class DataSet:
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    size: Optional[List]
    fileFormat: Optional[str]
    binaryOrText: Optional[str]
    proprietary: Optional[str]
    reference: Optional[List]
    toArchive: Optional[List]
    dataType: Optional[List[Relatant]] = field(default_factory=list)
    representationFormat: Optional[List[Relatant]] = field(default_factory=list)
    
    @classmethod
    def from_query(cls, raw_data: dict) -> 'DataSet':

        # Load MSC Classification
        options = get_options()

        data = raw_data[0]

        return cls(
            id = None,
            label = None,
            description = None,
            size = [] if '' in (temp := [options[data['sizeUnit']['value']] if data.get('sizeUnit', {}).get('value', '') else options['items'] if data.get('sizeRecord', {}).get('value', '') else '', data['sizeValue']['value'] if data.get('sizeValue', {}).get('value', '') else data['sizeRecord']['value'] if data.get('sizeRecord', {}).get('value', '') else '']) else temp,
            fileFormat = data.get('fileFormat', {}).get('value'),
            binaryOrText = options[data['binaryOrText']['value']] if data.get('binaryOrText', {}).get('value') else '',
            proprietary = options[data['proprietary']['value']] if data.get('proprietary', {}).get('value') else '',
            reference = ({order_to_publish()[data['publish']['value']][0]: [order_to_publish()[data['publish']['value']][1], '']} if data.get('publish', {}).get('value') else {}) | ({order_to_publish()['doi'][0]: [order_to_publish()['doi'][1], data['DOI']['value']]} if data.get('DOI', {}).get('value') else {}) | ({order_to_publish()['url'][0]: [order_to_publish()['url'][1], data['URL']['value']]} if data.get('URL', {}).get('value') else {}),
            toArchive = (temp := [options[data['archive']['value']] if data.get('archive', {}).get('value') else '',data.get('endTime', {}).get('value', '')[:4] if data.get('endTime', {}).get('value') else '']) and temp if temp[0] != '' else [],
            dataType = [Relatant.from_query(dtype) for dtype in data.get('dataType', {}).get('value', '').split(" / ") if dtype] if 'dataType' in data else [],
            representationFormat = [Relatant.from_query(rformat) for rformat in data.get('representationFormat', {}).get('value', '').split(" / ") if rformat] if 'representationFormat' in data else [] 
        )






from ..config import BASE_URI
from ..getters import get_options, get_questions

# URI PREFIX Map
def get_uri_prefix_map():
    questions = get_questions('workflow')
    URI_PREFIX_MAP = {
        f'{BASE_URI}{questions["Process Step"]["Input"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Data Set"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Data Set"]["ID"]["uri"]}',
            "prefix": "DS"
        },
        f'{BASE_URI}{questions["Process Step"]["Output"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Data Set"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Data Set"]["ID"]["uri"]}',
            "prefix": "DS"
        },
        f'{BASE_URI}{questions["Process Step"]["Method"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Method"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Method"]["ID"]["uri"]}',
            "prefix": "M"
        },
        f'{BASE_URI}{questions["Process Step"]["Environment-Software"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Software"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Software"]["ID"]["uri"]}',
            "prefix": "S"
        },
        f'{BASE_URI}{questions["Process Step"]["Environment-Instrument"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Instrument"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Instrument"]["ID"]["uri"]}',
            "prefix": "I"
        },
        f'{BASE_URI}{questions["Method"]["Software"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Software"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Software"]["ID"]["uri"]}',
            "prefix": "S"
        },
        f'{BASE_URI}{questions["Method"]["Instrument"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Instrument"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Instrument"]["ID"]["uri"]}',
            "prefix": "I"
        },
        f'{BASE_URI}{questions["Instrument"]["Software"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Software"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Software"]["ID"]["uri"]}',
            "prefix": "S"
        },
        f'{BASE_URI}{questions["Hardware"]["Software"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Software"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Software"]["ID"]["uri"]}',
            "prefix": "S"
        }
    }
    return URI_PREFIX_MAP


# Dictionary with list of property names
PROPS = {
    'PS2IDS': ['inputDataSet'],
    'PS2ODS': ['outputDataSet'],
    'PS2M': ['uses'],
    'PS2PLS': ['platformSoftware'],
    'PS2PLI': ['platformInstrument'],
    'PS2F': ['fieldOfWork'],
    'PS2MA': ['mscID'],
    'M2S': ['implementedBySoftware'],
    'M2I': ['implementedByInstrument'],
    'S2PL': ['programmedIn'],
    'S2DP': ['dependsOnSoftware'],
    'H2CPU': ['CPU'],
    'DS2DT': ['dataType'],
    'DS2RF': ['representationFormat']
}

# Order of toPublish Answers
def order_to_publish(): 
    options = get_options()
    order = {
        'Yes': (0, options['Yes']),
        'doi': (1, options['DOI']),
        'url': (2, options['URL']),
        'No': (3, options['URL'])
        }
    return order

#Dictionary For Reproducibility
REPRODUCIBILITY = {
    'mathematical': 'mathematically reproducible research workflow',
    'runtime': 'runtime reproducible research workflow',
    'result': 'result reproducible research workflow',
    'originalplatform': 'original platform reproducible research workflow',
    'otherplatform': 'cross-platform reproducible research workflow'
    }




'''Worker Module for Model Preview and Export'''

import logging
import time

from .utils import map_entity_quantity
from .constants import (
    preview_relations,
    preview_map_general,
    preview_map_quantity,
    get_relations,
    get_data_properties
)

from ..config import endpoint
from ..getters import get_items, get_mathmoddb, get_properties
from ..helpers import entity_relations, map_entity, unique_items, date_precision
from ..queries import query_sparql
from ..payload import GeneratePayload


class PrepareModel:
    '''Class preparing Model Answers for Preview and Export'''
    def __init__(self):
        self.mathmoddb = get_mathmoddb()
        self.items = get_items()
        self.properties = get_properties()

    def preview(self, answers):
        '''Function to establish relations between Model Documentation Data'''

        # Prepare Relations for Preview
        for relation in preview_relations:
            entity_relations(
                data = answers,
                idx = {
                    'from': relation['from_idx'],
                    'to': relation['to_idx']
                },
                entity = {
                    'relation': relation['relation'],
                    'old_name': relation['old_name'],
                    'new_name': relation['new_name'],
                    'encryption': relation['encryption']
                },
                order = {
                    'formulation': relation['formulation'],
                    'task': relation['task']
                }
            )

        # Prepare General Mappings
        for mapping in preview_map_general:
            map_entity(
                data = answers,
                idx = {
                    'from': mapping[0], 
                    'to': mapping[1]
                },
                entity = {
                    'old_name': mapping[2], 
                    'new_name': mapping[3], 
                    'encryption': mapping[4]
                }
            )

        # Prepare Quantity Mapping
        for mapping in preview_map_quantity:
            map_entity_quantity(
                data = answers,
                entity_type = mapping,
                mapping = self.mathmoddb)

        return answers

    def export(self, data, url):
        """Function to create Payload for Model Export."""

        items = unique_items(data)

        payload = GeneratePayload(
            url = url,
            items = items,
            relations = get_relations(),
            data_properties = get_data_properties,
        )

        # Add / Retrieve Components of Model Item
        payload.process_items()

        # Delegate to helper functions
        self._export_fields(
            payload = payload,
            fields = data.get("field", {}),
            problems = data.get("problem", {})
        )
        self._export_problems(
            payload = payload,
            models = data.get("model", {}),
            problems = data.get("problem", {})
        )
        self._export_models(
            payload = payload,
            models = data.get("model", {})
        )
        self._export_tasks(
            payload = payload,
            tasks = data.get("task", {})
        )
        self._export_formulations(
            payload = payload,
            formulations = data.get("formulation", {})
        )
        self._export_quantities(
            payload = payload,
            quantities = data.get("quantity", {})
        )
        self._export_publications(
            payload = payload,
            publications = data.get("publication", {})
        )

        # Construct Item Payloads
        payload.add_item_payload()

        # If Relations are added, check if they exist
        if any(
            key.startswith("RELATION")
            for key in payload.get_dictionary()
        ):
            query = payload.build_relation_check_query()
            check = None
            for attempt in range(2):  # try twice
                try:
                    check = query_sparql(
                        query = query,
                        sparql_endpoint = endpoint["mardi"]["sparql"]
                    )
                    break
                except Exception as e:
                    logging.warning("SPARQL query attempt %s failed: %s", attempt + 1, e)
                    if attempt == 0:
                        time.sleep(1)  # short wait before retry
            if check is None:
                # both attempts failed → pretend no results
                check = [{}]

            payload.add_check_results(
                check = check
            )

        return payload.get_dictionary()

    # ---------------------------
    # Shared helper
    # ---------------------------
    def _add_common_metadata(self, payload, qclass, profile_type):
        """Add metadata common to most entities (except publication)."""
        payload.add_answer(
            verb = self.properties["instance of"],
            object_and_type = [qclass, "wikibase-item"],
        )

        payload.add_answer(
            verb = self.properties["community"],
            object_and_type = [self.items["MathModDB"], "wikibase-item"],
        )

        payload.add_answer(
            verb = self.properties["MaRDI profile type"],
            object_and_type = [self.items[profile_type], "wikibase-item"],
        )

        payload.add_answers(
            mardmo_property = "descriptionLong",
            wikibase_property = "description",
        )

    # ---------------------------
    # Entity export helpers
    # ---------------------------
    def _export_fields(self, payload, fields: dict, problems: dict):
        for entry in fields.values():
            if not entry.get("ID"):
                continue

            payload.get_item_key(
                value = entry
            )

            self._add_common_metadata(
                payload = payload,
                qclass = self.items["academic discipline"],
                profile_type = "MaRDI research field profile",
            )

            payload.add_backward_relation(
                data = problems.values(),
                relation = self.properties["contains"],
                relatants = "RFRelatant",
            )

            payload.add_intra_class_relation(
                relation = "IntraClassRelation",
                relatant = "IntraClassElement",
            )

    def _export_problems(self, payload, models: dict, problems: dict):
        for entry in problems.values():
            if not entry.get("ID"):
                continue

            payload.get_item_key(
                value = entry
            )

            self._add_common_metadata(
                payload = payload,
                qclass = self.items["research problem"],
                profile_type = "MaRDI research problem profile",
            )

            payload.add_backward_relation(
                data = models.values(),
                relation = self.properties["modelled by"],
                relatants = "RPRelatant",
            )

            payload.add_intra_class_relation(
                relation = "IntraClassRelation",
                relatant = "IntraClassElement",
            )

    def _export_models(self, payload, models: dict):
        for entry in models.values():
            if not entry.get("ID"):
                continue

            payload.get_item_key(
                value = entry
            )

            self._add_common_metadata(
                payload = payload,
                qclass = self.items["mathematical model"],
                profile_type = "MaRDI model profile",
            )

            payload.add_data_properties(
                item_class = "model"
            )

            payload.add_forward_relation_multiple(
                relation = "MM2MF",
                relatant = "MFRelatant",
            )

            payload.add_forward_relation_single(
                relation = self.properties["used by"],
                relatant = "TRelatant",
            )

            payload.add_intra_class_relation(
                relation = "IntraClassRelation",
                relatant = "IntraClassElement",
            )

    def _export_tasks(self, payload, tasks: dict):
        for entry in tasks.values():
            if not entry.get("ID"):
                continue

            payload.get_item_key(
                value = entry
            )

            self._add_common_metadata(
                payload = payload,
                qclass = self.items["computational task"],
                profile_type = "MaRDI task profile",
            )

            payload.add_data_properties(
                item_class = "task"
            )

            payload.add_forward_relation_multiple(
                relation = "T2MF",
                relatant = "MFRelatant",
            )

            payload.add_forward_relation_multiple(
                relation = "T2Q",
                relatant = "QRelatant",
            )

            payload.add_intra_class_relation(
                relation = "IntraClassRelation",
                relatant = "IntraClassElement",
            )

    def _export_formulations(self, payload, formulations: dict):
        for entry in formulations.values():
            if not entry.get("ID"):
                continue

            payload.get_item_key(
                value = entry
            )

            self._add_common_metadata(
                payload = payload,
                qclass = self.items["mathematical expression"],
                profile_type = "MaRDI formula profile",
            )

            payload.add_data_properties(
                item_class = "equation"
            )

            payload.add_answers(
                mardmo_property = "Formula",
                wikibase_property = "defining formula",
                datatype = "math",
            )

            payload.add_in_defining_formula()

            payload.add_forward_relation_multiple(
                relation = "MF2MF",
                relatant = "MFRelatant",
            )

            payload.add_intra_class_relation(
                relation = "IntraClassRelation",
                relatant = "IntraClassElement",
            )

    def _export_quantities(self, payload, quantities: dict):
        for entry in quantities.values():
            if not entry.get("ID"):
                continue

            payload.get_item_key(
                value = entry
            )

            if entry.get("QorQK") == self.mathmoddb["Quantity"]:
                self._add_common_metadata(
                    payload = payload,
                    qclass = self.items["quantity"],
                    profile_type = "MaRDI quantity profile",
                )
                qtype = "quantity"

            elif entry.get("QorQK") == self.mathmoddb["QuantityKind"]:
                self._add_common_metadata(
                    payload = payload,
                    qclass = self.items["kind of quantity"],
                    profile_type = "MaRDI quantity profile",
                )
                qtype = "quantity kind"

            else:
                continue

            payload.add_data_properties(
                item_class = qtype
            )

            if (
                entry.get("reference")
                and qtype == "quantity kind"
            ):
                payload.add_answer(
                    verb = self.properties["QUDT quantity kind ID"],
                    object_and_type = [entry["reference"][0][1], "external-id"],
                )

            payload.add_answers(
                mardmo_property = "Formula",
                wikibase_property = "defining formula",
                datatype = "math",
            )

            payload.add_in_defining_formula()

            if qtype == "quantity":
                payload.add_intra_class_relation(
                    relation = "Q2Q",
                    relatant = "QRelatant",
                )

                payload.add_intra_class_relation(
                    relation = "Q2QK",
                    relatant = "QKRelatant",
                )

            else:
                payload.add_intra_class_relation(
                    relation = "QK2QK",
                    relatant = "QKRelatant",
                )

                payload.add_intra_class_relation(
                    relation = "QK2Q",
                    relatant = "QRelatant",
                )

    def _export_publications(self, payload, publications: dict):
        for entry in publications.values():
            if not entry.get("ID"):
                continue

            payload.get_item_key(
                value = entry
            )

            # Only add bibliographic statements for non-MaRDI / non-Wikidata items
            if "mardi" not in entry["ID"] and "wikidata" not in entry["ID"]:
                # class
                if entry.get("entrytype") == "scholarly article":
                    pclass = self.items["scholarly article"]
                else:
                    pclass = self.items["publication"]

                payload.add_answer(
                    verb = self.properties["instance of"],
                    object_and_type = [pclass, "wikibase-item"],
                )

                # bibliographic data
                if entry.get("title"):
                    payload.add_answer(
                        verb = self.properties["title"],
                        object_and_type = [
                            {"text": entry["title"], "language": "en"},
                            "monolingualtext",
                        ],
                    )

                if entry.get("volume"):
                    payload.add_answer(
                        verb = self.properties["volume"],
                        object_and_type = [entry["volume"], "string"],
                    )

                if entry.get("issue"):
                    payload.add_answer(
                        verb = self.properties["issue"],
                        object_and_type = [entry["issue"], "string"],
                    )

                if entry.get("page"):
                    payload.add_answer(
                        verb = self.properties["page(s)"],
                        object_and_type = [entry["page"], "string"],
                    )

                if entry.get("date"):
                    payload.add_answer(
                        verb = self.properties["publication date"],
                        object_and_type = [
                            {
                                "time": f"+{entry['date']}",
                                "precision": date_precision(
                                    date_str = entry['date']
                                ),
                                "calendarmodel": (
                                    "http://www.wikidata.org/entity/Q1985727"
                                ),
                            },
                            "time",
                        ],
                    )

                if entry.get("reference", {}).get(0):
                    payload.add_answer(
                        verb = self.properties["DOI"],
                        object_and_type = [
                            entry["reference"][0][1].upper(),
                            "external-id"
                        ],
                    )

                # relations
                payload.add_forward_relation_single(
                    relation = self.properties["language of work or name"],
                    relatant = "language",
                )
                payload.add_forward_relation_single(
                    relation = self.properties["published in"],
                    relatant = "journal",
                )
                payload.add_forward_relation_single(
                    relation = self.properties["author"],
                    relatant = "author",
                    alternative = {
                        "relation": self.properties["author name string"],
                        "relatant": "Name",
                    },
                )

                payload.add_answer(
                    verb = self.properties["MaRDI profile type"],
                    object_and_type = [
                        self.items["MaRDI publication profile"],
                        "wikibase-item"
                    ],
                )

            # Add relations to Entities of Mathematical Model
            payload.add_forward_relation_multiple(
                relation = "P2E",
                relatant = "EntityRelatant",
                reverse = True,
            )



'''Module containing Constants for the Model Documentation'''

from ..config import BASE_URI
from ..getters import get_items, get_mathmoddb, get_properties, get_questions

# Data Properties Check
data_properties_check = {
    ('is_deterministic', 'is_stochastic'): '(Deterministic and Stochastic)',
    ('is_dimensionless', 'is_dimensional'): '(Dimensionless and Dimensional)',
    ('is_dynamic', 'is_static'): '(Dynamic and Static)',
    ('is_linear', 'is_not_linear'): '(Linear and Not Linear)',
    ('is_space_continuous', 'is_space_discrete'): '(Space-Continuous and Space-Discrete)',
    ('is_time_continuous', 'is_time_discrete'): '(Time-Continuous and Time-Discrete)',
    ('is_mathematical_constant', 'is_physical_constant'): '(Mathematical and Physical Constant)',
    ('is_mathematical_constant', 'is_physical_constant'): '(Mathematical and Chemical Constant)'
}

# Data Properties per Class
data_properties_per_class = {
    'model': [
        'is_linear', 'is_not_linear', 'is_dynamic', 'is_static',
        'is_deterministic', 'is_stochastic', 'is_dimensionless', 'is_dimensional',
        'is_time_continuous', 'is_time_discrete', 'is_space_continuous', 'is_space_discrete',
    ],
    'quantity': [
        'is_chemical_constant', 'is_mathematical_constant', 'is_physical_constant',
        'is_dynamic', 'is_static', 'is_deterministic', 'is_stochastic', 'is_dimensionless',
        'is_dimensional', 'is_time_continuous', 'is_time_discrete', 'is_space_continuous',
        'is_space_discrete'
    ],
    'formulation': [
        'is_linear', 'is_not_linear', 'is_dynamic', 'is_static', 'is_deterministic',
        'is_stochastic', 'is_dimensionless', 'is_dimensional', 'is_time_continuous',
        'is_time_discrete', 'is_space_continuous', 'is_space_discrete'
    ],
    'task': ['is_linear', 'is_not_linear', 'is_time_continuous', 'is_time_discrete',
             'is_space_continuous', 'is_space_discrete'
    ]
}

# Dictionary for internal / external section names
section_map = {
    'model': 'Mathematical Model',
    'task': 'Computational Task',
    'formulation': 'Mathematical Expression',
    'quantity': 'Quantity [Kind]',
    'problem': 'Research Problem',
    'field': 'Academic Discipline',
    'publication': 'Publication'
}

# Dictionary with list of property names
props = {
    'T2MF': [
        'assumes',
        'contains_formulation',
        'contains_boundary_condition',
        'contains_constraint_condition',
        'contains_coupling_condition',
        'contains_initial_condition',
        'contains_final_condition'
    ],
    'T2Q': [
        'contains_input',
        'contains_output',
        'contains_objective',
        'contains_parameter',
        'contains_constant'
    ],
    'RP2RF': [
        'contained_in_field'
    ],
    'MM2RP': [
        'models'
    ],
    'MM2T': [
        'used_by'
    ],
    'MM2MF': [
        'assumes',
        'contains_formulation',
        'contains_boundary_condition',
        'contains_constraint_condition',
        'contains_coupling_condition',
        'contains_initial_condition',
        'contains_final_condition'
    ],
    'MF2MF': [
        'assumes',
        'contains_formulation',
        'contains_boundary_condition',
        'contains_constraint_condition',
        'contains_coupling_condition',
        'contains_initial_condition',
        'contains_final_condition'
    ],
    'Field': [
        "specialized_by",
        "specializes",
        "similar_to"
    ],
    'Problem': [
        "specialized_by",
        "specializes",
        "similar_to"
    ],
    'Model': [
        'specialized_by',
        'specializes',
        'discretized_by',
        'discretizes',
        'contained_in_model',
        'contains_model',
        'approximated_by',
        'approximates',
        'linearized_by',
        'linearizes',
        'similar_to'
    ],
    'Task': [
        'specialized_by',
        'specializes',
        'discretized_by',
        'discretizes',
        'contained_in_task',
        'contains_task',
        'approximated_by',
        'approximates',
        'linearized_by',
        'linearizes',
        'similar_to'
    ],
    'Formulation': [
        'specialized_by',
        'specializes',
        'discretized_by',
        'discretizes',
        'approximated_by',
        'approximates',
        'linearized_by',
        'linearizes',
        'nondimensionalized_by',
        'nondimensionalizes',
        'similar_to'
    ],
    'Quantity': [
        "specialized_by",
        "specializes",
        "approximated_by",
        "approximates",
        "discretized_by",
        "discretizes",
        "linearized_by",
        "linearizes",
        "nondimensionalized_by",
        "nondimensionalizes",
        "similar_to"
    ],
}

# Index counters for different qclass combinations
index_counters = {
    ("Quantity", "Quantity"): 0,
    ("QuantityKind", "QuantityKind"): 0,
    ("Quantity", "QuantityKind"): 0,
    ("QuantityKind", "Quantity"): 0,
}

# URI mappings for quantity relations
relation_uris = {
    ("Quantity", "Quantity"): "Q2Q",
    ("QuantityKind", "QuantityKind"): "QK2QK",
    ("Quantity", "QuantityKind"): "Q2QK",
    ("QuantityKind", "Quantity"): "QK2Q",
}

# URI mappings for quantity relatants
relatant_uris = {
    ("Quantity", "Quantity"): "QQRelatant",
    ("QuantityKind", "QuantityKind"): "QKQKRelatant",
    ("Quantity", "QuantityKind"): "QQKRelatant",
    ("QuantityKind", "Quantity"): "QKQRelatant",
}

# DATA Properties
def get_data_properties(item_type):
    '''Data Properties for the Model Documentation'''
    mathmoddb = get_mathmoddb()
    items = get_items()
    data_properties = {
        mathmoddb['is_chemical_constant']: items.get('chemical constant'),
        mathmoddb['is_mathematical_constant']: items.get('mathematical constant'),
        mathmoddb['is_physical_constant']: items.get('physical constant'),
        mathmoddb['is_deterministic']: items.get(f'deterministic {item_type}'),
        mathmoddb['is_stochastic']: items.get(f'probabilistic {item_type}'),
        mathmoddb['is_dimensional']: items.get(f'dimensional {item_type}'),
        mathmoddb['is_dimensionless']: items.get(f'dimensionless {item_type}'),
        mathmoddb['is_dynamic']: items.get(f'dynamic {item_type}'),
        mathmoddb['is_static']: items.get(f'static {item_type}'),
        mathmoddb['is_linear']: items.get(f'linear {item_type}'),
        mathmoddb['is_not_linear']: items.get(f'nonlinear {item_type}'),
        mathmoddb['is_space_continuous']: items.get(f'continuous-space {item_type}'),
        mathmoddb['is_space_discrete']: items.get(f'discrete-space {item_type}'),
        mathmoddb['is_time_continuous']: items.get(f'continuous-time {item_type}'),
        mathmoddb['is_time_discrete']: items.get(f'discrete-time {item_type}')
    }
    return data_properties

# Relations
def get_relations():
    '''Relations for the Model Documentation'''
    mathmoddb = get_mathmoddb()
    items = get_items()
    properties = get_properties()
    relations = {
        # Map MathModDB Relation on Wikibase Property + Qualifier Item
        mathmoddb['assumes']: [
            properties['assumes']
        ],
        mathmoddb['contains_boundary_condition']: [
            properties['contains'],
            items['boundary condition']
        ],
        mathmoddb['contains_constraint_condition']: [
            properties['contains'],
            items['constraint']
        ],
        mathmoddb['contains_coupling_condition']: [
            properties['contains'],
            items['coupling condition']
        ],
        mathmoddb['contains_initial_condition']: [
            properties['contains'],
            items['initial condition']
        ],
        mathmoddb['contains_final_condition']: [
            properties['contains'],
            items['final condition']
        ],
        mathmoddb['contains_input']: [
            properties['contains'],
            items['input']
        ],
        mathmoddb['contains_constant']: [
            properties['contains'],
            items['constant']
        ],
        mathmoddb['contains_objective']: [
            properties['contains'],
            items['objective function']
        ],
        mathmoddb['contains_output']: [
            properties['contains'],
            items['output']
        ],
        mathmoddb['contains_parameter']: [
            properties['contains'],
            items['parameter']
        ],
        mathmoddb['documents']: [
            properties['described by source'],
            items['documentation']
        ],
        mathmoddb['invents']: [
            properties['described by source'],
            items['invention']
        ],
        mathmoddb['studies']: [
            properties['described by source'],
            items['study']
        ],
        mathmoddb['surveys']: [
            properties['described by source'],
            items['review']
        ],
        mathmoddb['uses']: [
            properties['described by source'],
            items['use']
        ],
        # Map MathModDB Relation on Wikibase Property + Direction
        mathmoddb['specialized_by']: [
            properties['specialized by'],
            'forward'
        ],
        mathmoddb['specializes']: [
            properties['specialized by'],
            'backward'
        ],
        mathmoddb['approximated_by']: [
            properties['approximated by'],
            'forward'
        ],
        mathmoddb['approximates']: [
            properties['approximated by'],
            'backward'
        ],
        mathmoddb['discretized_by']: [
            properties['discretized by'],
            'forward'
        ],
        mathmoddb['discretizes']: [
            properties['discretized by'],
            'backward'
        ],
        mathmoddb['linearized_by']: [
            properties['linearized by'],
            'forward'
        ],
        mathmoddb['linearizes']: [
            properties['linearized by'],
            'backward'
        ],
        mathmoddb['nondimensionalized_by']: [
            properties['nondimensionalized by'],
            'forward'
        ],
        mathmoddb['nondimensionalizes']: [
            properties['nondimensionalized by'],
            'backward'
        ],
        mathmoddb['contains']: [
            properties['contains'],
            'forward'
        ],
        mathmoddb['contained_in']: [
            properties['contains'],
            'backward'
        ],
        mathmoddb['similar_to']: [
            properties['similar to'],
            'forward'
        ]
    }
    return relations

# Parameter for Entity relations
preview_relations = [
    {
        "from_idx": "model",
        "to_idx": "problem",
        "relation": None,
        "old_name": "RPRelatant",
        "new_name": "RelationRP",
        "encryption": "RP",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "model",
        "to_idx": "formulation",
        "relation": "MM2MF",
        "old_name": "MFRelatant",
        "new_name": "RelationMF",
        "encryption": "ME",
        "formulation": True,
        "task": False,
    },
    {
        "from_idx": "model",
        "to_idx": "task",
        "relation": None,
        "old_name": "TRelatant",
        "new_name": "RelationT",
        "encryption": "T",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "model",
        "to_idx": "model",
        "relation": "IntraClassRelation",
        "old_name": "IntraClassElement",
        "new_name": "RelationMM",
        "encryption": "MM",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "task",
        "to_idx": "formulation",
        "relation": "T2MF",
        "old_name": "MFRelatant",
        "new_name": "RelationMF",
        "encryption": "ME",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "task",
        "to_idx": "quantity",
        "relation": "T2Q",
        "old_name": "QRelatant",
        "new_name": "RelationQQK",
        "encryption": "QQK",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "task",
        "to_idx": "task",
        "relation": "IntraClassRelation",
        "old_name": "IntraClassElement",
        "new_name": "RelationT",
        "encryption": "T",
        "formulation": False,
        "task": True,
    },
    {
        "from_idx": "formulation",
        "to_idx": "formulation",
        "relation": "MF2MF",
        "old_name": "MFRelatant",
        "new_name": "RelationMF1",
        "encryption": "ME",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "formulation",
        "to_idx": "formulation",
        "relation": "IntraClassRelation",
        "old_name": "IntraClassElement",
        "new_name": "RelationMF2",
        "encryption": "ME",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "quantity",
        "to_idx": "quantity",
        "relation": "Q2Q",
        "old_name": "QRelatant",
        "new_name": "RelationQQ",
        "encryption": "QQK",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "quantity",
        "to_idx": "quantity",
        "relation": "QK2QK",
        "old_name": "QKRelatant",
        "new_name": "RelationQKQK",
        "encryption": "QQK",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "quantity",
        "to_idx": "quantity",
        "relation": "Q2QK",
        "old_name": "QKRelatant",
        "new_name": "RelationQQK",
        "encryption": "QQK",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "quantity",
        "to_idx": "quantity",
        "relation": "QK2Q",
        "old_name": "QRelatant",
        "new_name": "RelationQKQ",
        "encryption": "QQK",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "field",
        "to_idx": "field",
        "relation": "IntraClassRelation",
        "old_name": "IntraClassElement",
        "new_name": "RelationRF",
        "encryption": "AD",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "problem",
        "to_idx": "field",
        "relation": None,
        "old_name": "RFRelatant",
        "new_name": "RelationRF",
        "encryption": "AD",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "problem",
        "to_idx": "problem",
        "relation": "IntraClassRelation",
        "old_name": "IntraClassElement",
        "new_name": "RelationRP",
        "encryption": "RP",
        "formulation": False,
        "task": False,
    },
    {
        "from_idx": "publication",
        "to_idx": [
            "field", "problem", "model",
            "formulation", "quantity", "task",
        ],
        "relation": "P2E",
        "old_name": "EntityRelatant",
        "new_name": "RelationP",
        "encryption": ["RF", "RP", "MM", "ME", "QQK", "T"],
        "formulation": False,
        "task": False,
    },
]

preview_map_general = [
    # fromIDX, toIDX, entityOld, entityNew, enc
    ('model', 'formulation', 'assumption', 'assumptionMapped', 'ME'),
    ('task', 'formulation', 'assumption', 'assumptionMapped', 'ME'),
    ('formulation', 'formulation', 'assumption', 'assumptionMapped', 'ME'),
]

preview_map_quantity = [
    # type
    ('formulation'),
    ('quantity'),
]

# URI PREFIX Map (I)
def get_uri_prefix_map():
    '''URI Prefixes for the Model Documentation'''
    questions = get_questions('model')
    uri_prefix_map = {
        f'{BASE_URI}{questions["Task"]["QRelatant"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Quantity"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Quantity"]["ID"]["uri"]}',
            "prefix": "QQK"
        },
        f'{BASE_URI}{questions["Task"]["MFRelatant"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Mathematical Formulation"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Mathematical Formulation"]["ID"]["uri"]}',
            "prefix": "ME"
        },
        f'{BASE_URI}{questions["Mathematical Formulation"]["Element Quantity"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Quantity"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Quantity"]["ID"]["uri"]}',
            "prefix": "QQK"
        },
        f'{BASE_URI}{questions["Quantity"]["Element Quantity"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Quantity"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Quantity"]["ID"]["uri"]}',
            "prefix": "QQK"
        },
        f'{BASE_URI}{questions["Mathematical Model"]["MFRelatant"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Mathematical Formulation"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Mathematical Formulation"]["ID"]["uri"]}',
            "prefix": "ME"
        },
        f'{BASE_URI}{questions["Mathematical Model"]["Assumption"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Mathematical Formulation"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Mathematical Formulation"]["ID"]["uri"]}',
            "prefix": "ME"
        },
        f'{BASE_URI}{questions["Task"]["Assumption"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Mathematical Formulation"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Mathematical Formulation"]["ID"]["uri"]}',
            "prefix": "ME"
        },
        f'{BASE_URI}{questions["Mathematical Formulation"]["Assumption"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Mathematical Formulation"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Mathematical Formulation"]["ID"]["uri"]}',
            "prefix": "ME"
        },
        f'{BASE_URI}{questions["Mathematical Formulation"]["MFRelatant"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Mathematical Formulation"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Mathematical Formulation"]["ID"]["uri"]}',
            "prefix": "ME"
        },
        f'{BASE_URI}{questions["Research Problem"]["RFRelatant"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Research Field"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Research Field"]["ID"]["uri"]}',
            "prefix": "AD"
        },
        f'{BASE_URI}{questions["Mathematical Model"]["RPRelatant"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Research Problem"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Research Problem"]["ID"]["uri"]}',
            "prefix": "RP"
        },
        f'{BASE_URI}{questions["Mathematical Model"]["TRelatant"]["uri"]}': {
            "question_set": f'{BASE_URI}{questions["Task"]["uri"]}',
            "question_id": f'{BASE_URI}{questions["Task"]["ID"]["uri"]}',
            "prefix": "T"
        }
    }
    return uri_prefix_map



from ..config import BASE_URI, endpoint
from ..getters import get_items, get_options, get_properties, get_questions
from ..helpers import value_editor
from ..queries import query_sparql
from ..adders import add_basics, add_references, add_relations_static

from .constants import PROPS
from .models import Method, ProcessStep, Software, Hardware, DataSet
from .sparql import queryHandler

class Information:
    '''Class containing functions, querying external sources for specific
       entities and integrating the related metadata into the questionnaire.'''
    
    def __init__(self):
        # Load shared data once
        self.questions = get_questions("workflow")
        self.base = BASE_URI
        self.options = get_options()

    def software(self, instance):
        '''Software Information'''

        # Software specific Questions.
        software = self.questions["Software"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add Basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Software',
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, Id = instance.external_id.split(':')

        # Query source for further Information
        query = queryHandler[source]['software'].format(Id, **get_items(), **get_properties())
        results = query_sparql(query, endpoint[source]['sparql'])

        if not results:
            return
                    
        # Structure Results and load options
        data = Software.from_query(results)

        # Add References to Questionnaire
        add_references(
            project = instance.project,
            data = data,
            uri = f'{self.base}{software["Reference"]["uri"]}',
            set_prefix = instance.set_index
        )

        # Add Relations between Programming Language and Method to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['S2PL']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{software["Programming Language"]["uri"]}'
            }
        )

        # Add Relations between Programming Language and Method to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['S2DP']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{software["Dependency"]["uri"]}'
            }
        )

        # Software Source Code Published?
        if data.sourceCodeRepository:
            value_editor(
                project = instance.project, 
                uri = f'{self.base}{software["Published"]["uri"]}', 
                info = {
                    'text': data.sourceCodeRepository, 
                    'option': self.options['YesText'], 
                    'set_prefix': instance.set_index
                }
            )

        # Software User Manual Documented?
        if data.userManualURL: 
            value_editor(
                project = instance.project, 
                uri = f'{self.base}{software["Documented"]["uri"]}', 
                info = {
                    'text': data.userManualURL,
                    'option': self.options['YesText'],
                    'set_prefix': instance.set_index
                }
            )

    def hardware(self, instance):
        '''Hardware Information'''

        # Hardwre specific Questions.
        hardware = self.questions["Hardware"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add Basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Hardware',
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, Id = instance.external_id.split(':')

        # Query source for further Information
        query = queryHandler[source]['hardware'].format(Id, **get_items(), **get_properties())
        results = query_sparql(query, endpoint[source]['sparql'])

        if not results:
            return

        # Structure Results
        data = Hardware.from_query(results)

        # Add Relations between CPU and Hardware to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['H2CPU']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{hardware["CPU"]["uri"]}'
            }
        )

        # Number of Nodes
        if data.nodes:
            value_editor(
                project = instance.project, 
                uri = f'{self.base}{hardware["Nodes"]["uri"]}', 
                info = {
                    'text': data.nodes, 
                    'set_prefix': instance.set_index
                }
            )

        # Number of Cores
        if data.cores:
            value_editor(
                project = instance.project, 
                uri = f'{self.base}{hardware["Cores"]["uri"]}', 
                info = {
                    'text': data.cores, 
                    'set_prefix': instance.set_index
                }
            )                

    def instrument(self, instance):
        '''Instrument Information'''

        # Instrument specific Questions.
        instrument = self.questions["Instrument"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add Basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Instrument',
            index = (0, instance.set_index)
        )

    def data_set(self, instance):
        '''Data Set Information'''

        # Data Set specific Questions.
        data_set = self.questions["Data Set"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add Basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Data Set',
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, Id = instance.external_id.split(':')

        # Query source for further Information
        query = queryHandler[source]['data-set'].format(Id, **get_items(), **get_properties())
        results = query_sparql(query, endpoint[source]['sparql'])

        if not results:
            return

        # Structure Results and load Pptions
        data = DataSet.from_query(results)

        # Data Set Size
        if data.size: 
            value_editor(
                project = instance.project, 
                uri = f'{self.base}{data_set["Size"]["uri"]}', 
                info = {
                    'text': data.size[1],
                    'option': data.size[0],
                    'set_prefix': instance.set_index
                }
            )

        # Add Relations between Data Type and Data Set to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['DS2DT']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{data_set["Data Type"]["uri"]}'
            }
        )

        # Add Relations between Representation Format and Data Set to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['DS2RF']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{data_set["Representation Format"]["uri"]}'
            }
        )

        # File Format
        if data.fileFormat:
            value_editor(
                project = instance.project, 
                uri = f'{self.base}{data_set["File Format"]["uri"]}', 
                info = {
                    'text': data.fileFormat, 
                    'set_prefix': instance.set_index
                }
            )

        # Binary or Text
        if data.binaryOrText:
            value_editor(
                project = instance.project, 
                uri = f'{self.base}{data_set["Binary or Text"]["uri"]}', 
                info = {
                    'option': data.binaryOrText, 
                    'set_prefix': instance.set_index
                }
            )

        # Proprietary
        if data.proprietary:
            value_editor(
                project = instance.project, 
                uri = f'{self.base}{data_set["Proprietary"]["uri"]}', 
                info = {
                    'option': data.proprietary, 
                    'set_prefix':  instance.set_index
                }
            )

        # References To Publish
        add_references(
            project = instance.project,
            data = data,
            uri = f'{self.base}{data_set["To Publish"]["uri"]}',
            set_prefix = instance.set_index
        )

        # Archiving
        if data.toArchive:
            value_editor(
                project = instance.project, 
                uri = f'{self.base}{data_set["To Archive"]["uri"]}', 
                info = {
                    'text': data.toArchive[1],
                    'option': data.toArchive[0], 
                    'set_prefix': instance.set_index
                }
            )

    def method(self, instance):
        '''Method Information'''

        # Method specific Questions.
        method = self.questions["Method"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Method',
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, Id = instance.external_id.split(':')

        # Query source for further Information
        query = queryHandler[source]['method'].format(Id, **get_items(), **get_properties())
        results = query_sparql(query, endpoint[source]['sparql'])

        if not results:
            return

        # Structure Results
        data = Method.from_query(results)

        # Add Relations between Software and Method to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['M2S']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{method["Software"]["uri"]}'
            }
        )

        # Add Relations between Instrument and Method to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['M2I']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{method["Instrument"]["uri"]}'
            }
        )

        return

    def process_step(self, instance):
        '''Process Step Information'''

        # Process Step specific Questions.
        process_step = self.questions["Process Step"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Process Step',
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, Id = instance.external_id.split(':')

        # Query source for further Information
        query = queryHandler[source]['step'].format(Id, **get_items(), **get_properties())
        results = query_sparql(query, endpoint[source]['sparql'])

        if not results:
            return

        # Structure Results
        data = ProcessStep.from_query(results)

        # Add Relations between Input Data Set and Process Step to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['PS2IDS']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{process_step["Input"]["uri"]}'
            }
        )

        # Add Relations between Output Data Set and Process Step to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['PS2ODS']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{process_step["Output"]["uri"]}'
            }
        )

        # Add Relations between Method and Process Step to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['PS2M']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{process_step["Method"]["uri"]}'
            }
        )

        # Add Relations between Software Platform and Process Step to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['PS2PLS']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{process_step["Environment-Software"]["uri"]}'
            }
        )

        # Add Relations between Instrument Platform and Process Step to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['PS2PLI']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{process_step["Environment-Instrument"]["uri"]}'
            }
        )

        # Add Relations between Fields and Process Step to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['PS2F']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{process_step["Discipline"]["uri"]}'
            }
        )

        # Add Relations between Math Areas and Process Step to Questionnaire
        add_relations_static(
            project = instance.project, 
            data = data, 
            props = {
                'keys': PROPS['PS2MA']
            }, 
            index = {
                'set_prefix': instance.set_index
            }, 
            statement = {
                'relatant': f'{self.base}{process_step["Discipline"]["uri"]}'
            }
        )



'''Module containing Models for the Publication Documentation'''

from dataclasses import dataclass, field
from typing import Optional, ClassVar

from ..getters import get_items, get_options
from ..helpers import date_format, split_value
from ..models import Relatant, RelatantWithClass

@dataclass
class Author:
    '''Data Class For Authors'''
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    orcid_id: Optional[str]
    zbmath_id: Optional[str]
    wikidata_id: Optional[str]

    @classmethod
    def from_query(cls, raw: str) -> 'Author':
        '''Generate Class Item from Query'''
        parts = raw.split(" <|> ")
        return cls(
            id = parts[0],
            label = parts[1],
            description = parts[2],
            orcid_id = parts[3],
            zbmath_id = parts[4],
            wikidata_id = parts[5]
        )

    @classmethod
    def from_crossref(cls, raw: str) -> 'Author':
        '''Generate Class Item from Crossref'''
        # Get Label
        label = None
        if raw.get('given') and raw.get('family'):
            label = f"{raw['given']} {raw['family']}"

        # Get MaRDI Portal ID
        identifier = None
        if label:
            identifier = 'no author found'

        # Get ORCID ID
        orcid_id = None
        if raw.get('ORCID'):
            orcid_id = raw.get('ORCID').split('/')[-1]

        # Get ZBMath ID
        zbmath_id = None

        # Get Wikidata ID
        wikidata_id = None

        # Get Label
        label = None
        if raw.get('given') and raw.get('family'):
            label = f"{raw['given']} {raw['family']}"

        # Get Description
        description = None
        if label:
            if orcid_id:
                description = f'scientist (ORCID iD {orcid_id})'
            else:
                description = 'scientist'


        return cls(
            id = identifier,
            label = label,
            description = description,
            orcid_id = orcid_id,
            zbmath_id = zbmath_id,
            wikidata_id = wikidata_id
        )

    @classmethod
    def from_datacite(cls, raw: str) -> 'Author':
        '''Generate Class Item from Datacite'''
        # Get Label
        label = None
        if raw.get('givenName') and raw.get('familyName'):
            label = f"{raw['givenName']} {raw['familyName']}"

        # Get MaRDI Portal ID
        identifier = None
        if label:
            identifier = 'no author found'

        # Get ORCID ID
        orcid_id = None
        for name_identifier in raw.get('nameIdentifiers', []):
            if name_identifier.get('nameIdentifierScheme') == 'ORCID':
                orcid_uri = name_identifier.get('nameIdentifier', '')
                orcid_id = orcid_uri.split('/')[-1]
                break

        # Get ZBMath ID
        zbmath_id = None

        # Get Wikidata ID
        wikidata_id = None

        # Get Description
        description = None
        if label:
            if orcid_id:
                description = f'scientist (ORCID iD {orcid_id})'
            else:
                description = 'scientist'

        return cls(
            id = identifier,
            label = label,
            description = description,
            orcid_id = orcid_id,
            zbmath_id = zbmath_id,
            wikidata_id = wikidata_id
        )

    @classmethod
    def from_doi(cls, raw: str) -> 'Author':
        '''Generate Class Item from DOI'''
        # Get Label
        label = None
        if raw.get('given') and raw.get('family'):
            label = f"{raw['given']} {raw['family']}"

        # Get MaRDI Portal ID
        identifier = None
        if label:
            identifier = 'no author found'

        # Get ORCID ID
        orcid_id = None
        if raw.get('ORCID'):
            orcid_id = raw.get('ORCID').split('/')[-1]

        # Get ZBMath ID
        zbmath_id = None

        # Get Wikidata ID
        wikidata_id = None

        # Get Description
        description = None
        if label:
            if orcid_id:
                description = f'scientist (ORCID iD {orcid_id})'
            else:
                description = 'scientist'

        return cls(
            id = identifier,
            label = label,
            description = description,
            orcid_id = orcid_id,
            zbmath_id = zbmath_id,
            wikidata_id = wikidata_id
        )

    @classmethod
    def from_zbmath(cls, raw: str) -> 'Author':
        '''Generate Class Item from zbMath'''
        # Get Label
        label = None
        if raw.get('name'):
            label = " ".join(reversed(raw['name'].split(", ")))

        # Get MaRDI Portal ID
        identifier = None
        if label:
            identifier = 'no author found'

        # Get ORCID ID
        orcid_id = None

        # Get zbMath ID
        zbmath_id = None
        if raw.get('codes'):
            zbmath_id = raw['codes'][0]

        # Get Wikidata ID
        wikidata_id = None

        # Get Description
        description = None
        if label:
            if zbmath_id:
                description = f'scientist (zbMath ID {zbmath_id})'
            else:
                description = 'scientist'

        return cls(
            id = identifier,
            label = label,
            description = description,
            orcid_id = orcid_id,
            zbmath_id = zbmath_id,
            wikidata_id = wikidata_id
        )

    @classmethod
    def from_orcid(cls, raw: str) -> 'Author':
        '''Generate Class Item from ORCiD'''
        # Get Label
        label = None

        # Get Name
        name = raw.get('name', {})
        given = name.get('given-names', {}).get('value')
        family = name.get('family-name', '').get('value')

        if given and family:
            label = f"{given} {family}"

        # Get MaRDI Portal ID
        identifier = None
        if label:
            identifier = 'no author found'

        # Get ORCID ID
        orcid_id = raw.get('name', {}).get('path')

        # Get ZBMath ID
        zbmath_id = None

        # Get Wikidata ID
        wikidata_id = None

        # Get Description
        description = None
        if label:
            if orcid_id:
                description = f'scientist (ORCID iD {orcid_id})'
            else:
                description = 'scientist'

        return cls(
            id = identifier,
            label = label,
            description = description,
            orcid_id = orcid_id,
            zbmath_id = zbmath_id,
            wikidata_id = wikidata_id
        )

@dataclass
class Journal:
    '''Data Class For Journals'''
    id: str
    issn: str
    label: str
    description: str

    @classmethod
    def from_query(cls, raw: str) -> 'Journal':
        '''Generate Class Item from Query'''
        identifier, label, description = raw.split(" <|> ")
        return cls(
            id = identifier,
            issn =None,
            label = label,
            description = description
        )

    @classmethod
    def from_crossref(cls, ids: list, item: list) -> 'Journal':
        '''Generate Class Item from Crossref'''
        # Get Label
        label = None
        if item:
            label = item[0]

        # Get ISSN
        issn = None
        if ids:
            issn = ids[0]

        # Get MaRDI Portal ID
        identifier = None
        if label:
            identifier = 'no journal found'

        # Get Description
        description = None
        if label:
            if issn:
                description = f'scientific journal (ISSN {issn})'
            else:
                description = 'scientific journal'

        return cls(
            id = identifier,
            issn = issn,
            label = label,
            description = description,
        )

    @classmethod
    def from_datacite(cls, ids: list, item: list) -> 'Journal':
        '''Generate Class Item from Datacite'''
        # Get Label
        label = None
        if item:
            label = (item[0].get('titles') or [{}])[0].get('title')

        # Get ISSN
        issn = None
        if ids:
            if ids[0].get('relatedIdentifierType') == 'ISSN':
                issn = ids[0].get('relatedIdentifier')

        # Get MaRDI Portal ID
        identifier = None
        if label:
            identifier = 'no journal found'

        # Get Description
        description = None
        if label:
            if issn:
                description = f'scientific journal (ISSN {issn})'
            else:
                description = 'scientific journal'

        return cls(
            id = identifier,
            issn = issn,
            label = label,
            description = description,
        )

    @classmethod
    def from_doi(cls, ids: list, item: str) -> 'Journal':
        '''Generate Class Item from DOI'''
        # Get Label
        label = None
        if item:
            label = item

        # Get ISSN
        issn = None
        if ids:
            issn = ids[0]

        # Get MaRDI Portal ID
        identifier = None
        if label:
            identifier = 'no journal found'

        # Get Description
        description = None
        if label:
            if issn:
                description = f'scientific journal (ISSN {issn})'
            else:
                description = 'scientific journal'

        return cls(
            id = identifier,
            issn = issn,
            label = label,
            description = description,
        )

    @classmethod
    def from_zbmath(cls, raw: dict) -> 'Journal':
        '''Generate Class Item from zbMath'''
        series = (raw.get('series') or [{}])[0]

        # Get Label
        label = None
        if series.get('title'):
            label = series['title']

        # Get ISSN
        issn = None
        if (series.get('issn') or [{}])[0].get('number'):
            issn = series['issn'][0]['number']

        # Get MaRDI Portal ID
        identifier = None
        if label:
            identifier = 'no journal found'

        # Get Description
        description = None
        if label:
            if issn:
                description = f'scientific journal (ISSN {issn})'
            else:
                description = 'scientific journal'

        return cls(
            id = identifier,
            issn = issn,
            label = label,
            description = description,
        )


@dataclass
class Publication:
    '''Data Class For Publications'''
    id: Optional[str] = None
    label: Optional[str] = None
    description: Optional[str] = None
    entrytype: Optional[str] = None
    title: Optional[str] = None
    date: Optional[str] = None
    volume: Optional[str] = None
    issue: Optional[str] = None
    page: Optional[str] = None
    reference: Optional[str] = None
    journal: list[Journal] = field(default_factory=list)
    authors: list[Author] = field(default_factory=list)
    language: list[Relatant] = field(default_factory=list)
    applies: list[RelatantWithClass] = field(default_factory=list)
    analyzes: list[RelatantWithClass] = field(default_factory=list)
    documents: list[RelatantWithClass] = field(default_factory=list)
    invents: list[RelatantWithClass] = field(default_factory=list)
    studies: list[RelatantWithClass] = field(default_factory=list)
    surveys: list[RelatantWithClass] = field(default_factory=list)
    uses: list[RelatantWithClass] = field(default_factory=list)

    options: ClassVar[Optional[dict]] = None

    @classmethod
    def get_options(cls) -> dict:
        '''Get Options for Generators'''
        if cls.options is None:
            cls.options = get_options()
        return cls.options

    @classmethod
    def from_query(cls, raw_data: dict) -> 'Publication':
        '''Generate Class Item from Query'''
        options = cls.get_options()
        items = get_items()

        data = raw_data[0]

        publication = {
            # Get ID
            'id': data.get('id', {}).get('value'),
            # Get Label
            'label': data.get('label', {}).get('value'),
            # Get Description
            'description': data.get('description', {}).get('value'),
            # Get Entrytype
            'entrytype': data.get('entrytypelabel', {}).get('value'),
            # Get Language
            'language': (
                [
                    Relatant.from_triple(
                        f"mardi:{items['english']}",
                        "English",
                        "West Germanic language"
                    )
                ]
                if data.get('langaugelabel', {}).get('value', '').lower()
                    in {"en", "eng", "english"}
                else []
            ),
            # Get Title
            'title': data.get('title', {}).get('value'),
            # Get Date
            'date': data.get('date', {}).get('value', '')[:10],
            # Get Volume
            'volume': data.get('volume', {}).get('value'),
            # Get Issue
            'issue': data.get('issue', {}).get('value'),
            # Get Page
            'page': data.get('page', {}).get('value'),
            # Get Reference
            'reference': {
                idx: [options['DOI'], data[prop]['value']]
                for idx, prop in enumerate(['doi'])
                if data.get(prop, {}).get('value')
            },
            # Get Journal
            'journal': 
                [
                    Journal.from_query(
                        data.get('journalInfo', {}).get('value')
                    )
                ]
                if 'journalInfo' in data else [],
            # Get Authors
            'authors': 
                [
                    Author.from_query(
                        author
                    )
                    for author in data.get('authorInfos', {}).get('value', '').split(" | ")
                    if author
                ]
                if 'authorInfos' in data else [],
            # Get Applies Relation(s)
            'applies': split_value(
                data = data,
                key = 'applies',
                transform = RelatantWithClass.from_query
            ),
            # Get Analyzes Relation(s)
            'analyzes': split_value(
                data = data,
                key = 'analyzes',
                transform = RelatantWithClass.from_query
            ),
            # Get Documents Relation(s)
            'documents': split_value(
                data = data,
                key = 'documents',
                transform = RelatantWithClass.from_query
            ),
            # Get Invents Relation(s)
            'invents': split_value(
                data = data,
                key = 'invents',
                transform = RelatantWithClass.from_query
            ),
            # Get Studies Relation(s)
            'studies': split_value(
                data = data,
                key = 'studies',
                transform = RelatantWithClass.from_query
            ),
            # Get Surveys Relation(s)
            'surveys': split_value(
                data = data,
                key = 'surveys',
                transform = RelatantWithClass.from_query
            ),
            # Get Uses Relation(s)
            'uses': split_value(
                data = data,
                key = 'uses',
                transform = RelatantWithClass.from_query
            ),
        }

        return cls(
            **publication
            )

    @classmethod
    def from_crossref(cls, raw_data: dict) -> 'Publication':
        '''Generate Class Item from Crossref'''
        options = cls.get_options()
        items = get_items()

        data = raw_data.json().get('message', {})

        publication = {
            # Get ID
            'id': None,
            # Get Label
            'label': None,
            # Get Description
            'description': 
                f'scientific article (doi: {data["DOI"]})'
                if data.get("DOI") else 'scientific article',
            # Get Entrytype
            'entrytype':
                'scholarly article'
                if data.get('type') == 'journal-article' else 'publication',
            # Get Language
            'language': (
                [
                    Relatant.from_triple(
                        f"mardi:{items['english']}",
                        "English",
                        "West Germanic language"
                    )
                ]
                if (data.get('language') or '').lower()
                    in {"en", "eng", "english"}
                else []
            ),
            # Get Title
            'title': data.get('title', [''])[0],
            # Get Date
            'date': date_format(
                data.get('published', {}).get('date-parts', [[]])[0]
            ),
            # Get Volume
            'volume': data.get('volume'),
            # Get Issue
            'issue': data.get('issue'),
            # Get Page
            'page': data.get('page'),
            # Get Reference
            'reference': {
                idx: [options['DOI'], data[prop]]
                for idx, prop in enumerate(['DOI'])
                if data.get(prop, {})
            },
            # Get Journal
            'journal':
                [
                    Journal.from_crossref(
                        data.get('ISSN', []),
                        data.get('container-title', [])
                    )
                ]
                if 'ISSN' in data or 'container-title' in data else [],
            # Get Authors
            'authors': 
                [
                    Author.from_crossref(
                        author
                    )
                    for author in data.get('author', [])
                    if author
                ]
                if 'author' in data else [],
        }

        return cls(
            **publication
        )

    @classmethod
    def from_datacite(cls, raw_data: dict) -> 'Publication':
        '''Generate Class Item from Datacite'''
        options = cls.get_options()
        items = get_items()

        data = raw_data.json().get('data', {}).get('attributes', {})

        publication = {
            # Get ID
            'id': None,
            # Get Label
            'label': None,
            # Get Description
            'description': 
                f'scientific article (doi: {data["doi"]})'
                if data.get("doi") else 'scientific article',
            # Get Entrytype
            'entrytype':
                'scholarly article'
                if data.get('types', {}).get('bibtex') == 'article' else 'publication',
            # Get Language
            'language': (
                [
                    Relatant.from_triple(
                        f"mardi:{items['english']}",
                        "English",
                        "West Germanic language"
                    )
                ]
                if (data.get('language') or '').lower()
                    in {"en", "eng", "english"}
                else []
            ),
            # Get Title
            'title': (data.get('titles') or [{}])[0].get('title'),
            # Get Date
            'date': 
                next(
                    (
                        date_format(date_part['date'].split('-'))
                        for date_part in data.get('dates', [])
                        if date_part.get('dateType') == 'Issued' and date_part.get('date')
                    ),
                    None,
                ),
            # Get Volume
            'volume': (data.get('relatedItems') or [{}])[0].get('volume'),
            # Get Issue
            'issue': (data.get('relatedItems') or [{}])[0].get('issue'),
            # Get Page
            'page':
                (
                    f"{data['relatedItems'][0].get('firstPage', '')}-"
                    f"{data['relatedItems'][0].get('lastPage', '')}"
                )
                if data.get('relatedItems') else None,
            # Get Reference
            'reference': {
                idx: [options['DOI'], data[prop]]
                for idx, prop in enumerate(['doi'])
                if data.get(prop, {})
            },
            # Get Journal
            'journal': 
                [
                    Journal.from_datacite(
                        data.get('relatedIdentifiers') or [{}],
                        data.get('relatedItems') or [{}],
                    )
                ]
                if 'relatedIdentifiers' in data or 'relatedItems' in data else [],
            # Get Authors
            'authors':
                [
                    Author.from_datacite(author)
                    for author in data.get('creators', [])
                    if author
                ]
                if 'creators' in data else [],
        }

        return cls(
            **publication
        )

    @classmethod
    def from_doi(cls, raw_data: dict) -> 'Publication':
        '''Generate Class Item from DOI'''
        options = cls.get_options()
        items = get_items()

        data = raw_data.json()

        publication = {
            # Get ID
            'id': None,
            # Get Label
            'label': None,
            # Get Description
            'description': 
                f'scientific article (doi: {data["DOI"]})'
                if data.get("DOI") else 'scientific article',
            # Get Entrytype
            'entrytype': (
                'scholarly article'
                if data.get('type') in ('journal-article', 'article-journal')
                else 'publication'
            ),
            # Get Language
            'language': (
                [
                    Relatant.from_triple(
                        f"mardi:{items['english']}",
                        "English",
                        "West Germanic language"
                    )
                ]
                if (data.get('language') or '').lower()
                    in {"en", "eng", "english"}
                else []
            ),
            # Get Title
            'title': data.get('title'),
            # Get Date
            'date': date_format(
                data.get('published', {}).get('date-parts', [[]])[0]
            ),
            # Get Volume
            'volume': data.get('volume'),
            # Get Issue
            'issue': data.get('issue'),
            # Get Page
            'page': data.get('page'),
            # Get Reference
            'reference': {
                idx: [options['DOI'], data[prop]]
                for idx, prop in enumerate(['DOI'])
                if data.get(prop, {})
            },
            # Get Journal
            'journal':
                [
                    Journal.from_doi(
                        data.get('ISSN', []),
                        data.get('container-title', [])
                    )
                ]
                if 'ISSN' in data or 'container-title' in data else [],
            # Get Authors
            'authors':
                [
                    Author.from_doi(author)
                    for author in data.get('author', [])
                    if author
                ]
                if 'author' in data else [],
        }

        return cls(
            **publication
        )

    @classmethod
    def from_zbmath(cls, raw_data: dict) -> 'Publication':
        '''Generate Class Item from zbMath'''
        options = cls.get_options()
        items = get_items()

        data = raw_data.json().get('result', [''])[0]

        publication = {
            # Get ID
            'id': None,
            # Get Label
            'label': None,
            # Get Description
            'description': (
                f"scientific article (doi: {doi})"
                if (doi := next(
                    (
                        link.get("identifier")
                        for link in data.get("links", [])
                        if link.get("type") == "doi"
                    ),
                    None,
                ))
                else "scientific article"
            ),
            # Get Entrytype
            'entrytype': (
                'scholarly article'
                if data.get('document_type', {}).get('description') == 'journal article'
                else 'publication'
            ),
            # Get Language
            'language':(
                [
                    Relatant.from_triple(
                        f"mardi:{items['english']}",
                        "English",
                        "West Germanic language"
                    )
                ]
                if any(
                    code.lower() in {"en", "eng", "english"}
                    for code in (data.get("language", {}).get("languages") or [""])
                )
                else []
            ),
            # Get Date
            'date':
                f"{int(data['year']):04d}-00-00T00:00:00Z"
                if data.get('year') else None,
            # Get Title
            'title': data.get('title', {}).get('title'),
            # Get Volume
            'volume': (data.get('source', {}).get('series') or [''])[0].get('volume'),
            # Get Issue
            'issue': (data.get('source', {}).get('series') or [''])[0].get('issue'),
            # Get Page
            'page': data.get('source', {}).get('page'),
            # Get Reference
            'reference': (
                {0: [options['DOI'], doi]}
                if (
                    doi := next(
                    (
                        link.get('identifier')
                        for link in data.get('links', [])
                        if link.get('type') == 'doi'
                    ),
                    None
                    )
                )
                else {}
            ),
            # Get Journal
            'journal': 
                [
                    Journal.from_zbmath(
                        data.get('source', {})
                    )
                ]
                if 'source' in data else [],
            # Get Authors
            'authors':
                [
                    Author.from_zbmath(author)
                    for author in data.get('contributors', {}).get('authors', [])
                    if author
                ]
                if 'contributors' in data else [],
        }

        return cls(
            **publication
        )



'''Function that query Data'''

import logging
from multiprocessing.pool import ThreadPool

import requests
from rdmo.domain.models import Attribute

from .config import BASE_URI, endpoint
from .getters import get_sparql_query
from .helpers import extract_parts

logger = logging.getLogger(__name__)

def query_item(label, description, api=endpoint['mardi']['api']):
    '''API request returning an Item whose label or alias matches,
       and with matching description.'''
    # Only check Items with description
    if not description or description == 'No Description Provided!':
        return None

    # Get data from API
    data = query_api(api, label)

    # Normalize input label for comparison (case insensitive)
    norm_label = label.strip().lower()

    matched_items = []
    for item in data:
        item_label = item.get('label', '').strip().lower()
        item_description = item.get('description', '').strip()
        item_aliases = [alias.strip().lower() for alias in item.get('aliases', [])]

        # Check label or alias match AND description match
        if (
            (item_label == norm_label or norm_label in item_aliases)
            and item_description == description
        ):
            matched_items.append(item)

    if matched_items:
        return matched_items[0]['id']

    return None

def query_api(api_url, search_term, timeout=5):
    """API requests returning all Items with matching label."""
    try:
        response = requests.get(
            api_url,
            params={
                'action': 'wbsearchentities',
                'format': 'json',
                'language': 'en',
                'type': 'item',
                'limit': 10,
                'search': search_term
            },
            headers={'User-Agent': 'MaRDMO (https://zib.de; reidelbach@zib.de)'},
            timeout=timeout
        )
        response.raise_for_status()  # Raise an error on bad HTTP status codes
        try:
            return response.json().get('search', [])
        except ValueError:
            # Malformed JSON
            logger.error("Failed to parse JSON.")
    except requests.exceptions.RequestException as e:
        logger.error("Request failed due to %s", e)

    return []

def query_sparql(query, sparql_endpoint):
    '''SPARQL request returning all Items with matching properties.'''
    if not sparql_endpoint:
        logger.warning("SPARQL query attempted without a valid endpoint.")
        return []

    try:
        response = requests.post(
            sparql_endpoint,
            data=query.encode("utf-8"),
            headers={
                "User-Agent": "MaRDMO (https://zib.de; reidelbach@zib.de)",
                "Content-Type": "application/sparql-query; charset=UTF-8",
                "Accept": "application/sparql-results+json"
            },
            timeout = 60
        )
        # Check if request was successful
        if response.status_code == 200:
            return response.json().get('results', {}).get('bindings', [])

        logger.error(
            "SPARQL request to %s failed with status %s: %s",
            sparql_endpoint,
            response.status_code,
            response.text,
        )

        return []

    except requests.exceptions.ConnectionError:
        logger.error(
            "SPARQL query failed: Unable to connect to the %s.",
            sparql_endpoint
        )

    except requests.exceptions.RequestException as e:
        logger.exception("SPARQL request failed: %s", e)

    return []

def query_sparql_pool(query_input):
    '''Pooled SPARQL request returning all items with matching properties 
       from different endpoints'''
    pool = ThreadPool(processes = len(query_input))
    # Map each endpoint's query and store results in a dictionary
    results = pool.map(lambda args: query_sparql(*args), query_input.values())
    data = dict(zip(query_input.keys(), results))
    return data

def query_sources(search, query_id = '', sources = None, not_found = True):
    '''Helper function to query specified sources and process results.'''

    if sources is None:
        # Set default Sources
        sources = ['mardi', 'wikidata']

    def process_result(result, source):
        '''Function to process the result and return a dictionary with id and text'''
        display = result['display']
        label = display.get('label', {}).get('value', 'No Label Provided!')
        description = display.get('description', {}).get('value', 'No Description Provided!')
        return {
            'id': f"{source}:{result['id']}",
            'text': f"{label} ({description}) [{source}]"        
        }

    source_functions = {
        'wikidata': lambda s: query_api(endpoint['wikidata']['api'], s),
        'mardi': lambda s: query_api(endpoint['mardi']['api'], s),
        'mathalgodb': lambda s: query_provider(s, get_sparql_query(query_id), 'mathalgodb')
    }

    # Filter only specified sources
    queries = [source_functions[source] for source in sources if source in source_functions]

    # Use ThreadPool to make concurrent API requests
    pool = ThreadPool(processes=len(queries))
    results = pool.map(lambda func: func(search), queries)

    # Unpack results based on available sources
    results_dict = dict(zip(sources, results))

    # Process results to fit RDMO Provider Output Requirements
    options = []

    if 'mathalgodb' in results_dict:
        options += results_dict['mathalgodb'][:15]

    if 'mardi' in results_dict:
        options += [process_result(result, 'mardi') for result in results_dict['mardi'][:15]]

    if 'wikidata' in results_dict:
        options += [process_result(result, 'wikidata') for result in results_dict['wikidata'][:15]]

    if not_found:
        options = [{'id': 'not found', 'text': 'not found'}] + options

    return options

def query_provider(search, query, source):
    """
    Dynamic query of MathModDB and MathAlgoDB, results as options for Provider.
    """
    if not search:
        return []

    # Fetch results from the MathAlgoDB knowledge graph
    results = query_sparql(
        query = query,
        sparql_endpoint = endpoint['mathalgodb']['sparql']
    )

    dic = {}
    options = []

    # Store results in dict
    for result in results:
        dic.update(
            {
                result['label']['value']: {
                    'id': result['id']['value'],
                    'quote': result['quote']['value']
                }
            }
        )

    # Filter results by user-defined search
    options.extend(
        [
            {
                'id': f"{source}:{value['id']}",
                'text': f'{key} ({value["quote"]}) [{source}]'
            }
            for key, value in dic.items()
            if search.lower() in key.lower()
        ]
    )

    return options

def query_sources_with_user_additions(search, project, setup):
    '''Fetch options from KG, user-defined fields, and other sources.'''

    if setup['sources'] is None:
        setup['sources'] = ['mardi', 'wikidata']

    # Query sources and get the results directly in options
    try:
        options = query_sources(
            search = search,
            query_id = setup['query_id'],
            sources = setup['sources'],
            not_found = False
        )
    except(requests.exceptions.RequestException, KeyError, ValueError) as e:
        logger.error("Query sources failed: %s", e)
        options = []

    # Dictionaries for User Answers
    values = {}
    dic = {}

    for query_attribute in setup['query_attributes']:

        # Fetch User entries from the project (ID)
        values['id'] = project.values.filter(
            snapshot = None,
            attribute = Attribute.objects.get(
                uri = f'{BASE_URI}domain/{query_attribute}/id'
            )
        )

        # Fetch User entries from the project (Name)
        values['name'] = project.values.filter(
            snapshot = None,
            attribute=Attribute.objects.get(
                uri = f'{BASE_URI}domain/{query_attribute}/name'
            )
        )

        # Fetch User entries from the project (Description)
        values['description'] = project.values.filter(
            snapshot = None,
            attribute=Attribute.objects.get(
                uri = f'{BASE_URI}domain/{query_attribute}/description'
            )
        )

        # Zip user-defined answers
        zipped = zip(
            values['id'],
            values['name'],
            values['description']
        )

        # Process user-defined answers
        for idx, (value1, value2, value3) in enumerate(zipped):
            item = dict.fromkeys(["source", "label", "description", "id"], None)
            if value1.text:
                if value1.text == 'not found':
                    # User-Defined Cases
                    item['label'] = value2.text or "No Label Provided!"
                    item['description'] = value3.text or "No Description Provided!"
                    item['id'] = idx
                    item['source'] = 'user'
                else:
                    # ID Cases
                    item['label'], item['description'], item['source'] = extract_parts(value1.text)
                    _, item['id'] = value1.external_id.split(':')
            if item['source'] not in setup['sources']:
                if item['source'] == 'user':
                    dic[f"{item['label']} ({item['description']}) [{item['source']}]"] = {
                        'id': "not found"
                    }
                else:
                    dic[f"{item['label']} ({item['description']}) [{item['source']}]"] = {
                        'id': f"{item['source']}:{item['id']}"
                    }

    # Add the user-defined options to the list, filtered by search
    options.extend(
        [
            {
                'id': f"{value['id']}",
                'text': key
            }
            for key, value in dic.items() if search.lower() in key.lower()
        ]
    )

    if setup['creation']:
        options = [{'id': 'not found', 'text': f"{search}"}] + options

    # Return combined, sorted options
    return options #sorted(options, key=lambda option: option['text'])



"""
MaRDMO: A plugin for RDMO to document and search 
mathematical models, algorithms, and interdisciplinary 
workflows.

Modules:
- models: document mathematical models
- algorithms: document algorithms
- workflows: document interdisciplinary workflows
- publication: document publications for models, algorithms, and workflows
- search: search models, algorithms, and workflows
"""

__title__ = 'MaRDMO'
__version__ = '0.3.4'
__author__ = 'Marco Reidelbach'
__email__ = 'reidelbach@zib.de'
__license__ = 'Apache-2.0'

VERSION = __version__



'''Worker Module to collect Algorithm Metadata'''

from .constants import PREVIEW_RELATIONS

from ..helpers import entity_relations

def algorithm_relations(answers):
    '''Function to establish relations between Algorithm Documentation Data'''

    # Prepare Relations for Preview
    for relation in PREVIEW_RELATIONS:
        entity_relations(
            data = answers,
            idx = {
                'from': relation[0],
                'to': relation[1]
            },
            entity = {
                'relation': relation[2],
                'old_name': relation[3],
                'new_name': relation[4],
                'encryption': relation[5]
            },
            order = {
                'formulation': False,
                'task': False
            }
        )

    return answers



'''Functions to add Information to the Questionaire.'''

import json
import os

from django.apps import apps
from rdmo.domain.models import Attribute

from .config import BASE_URI, endpoint
from .constants import flag_dict
from .helpers import nested_set

def get_mathmoddb():
    """Retrieve the mathmoddb ontology from MaRDMOConfig."""
    return apps.get_app_config("MaRDMO").mathmoddb

def get_mathalgodb():
    """Retrieve the mathmoddb ontology from MaRDMOConfig."""
    return apps.get_app_config("MaRDMO").mathalgodb

def get_options():
    """Retrieve the rdmo options from MaRDMOConfig."""
    return apps.get_app_config("MaRDMO").options

def get_items():
    """Retrieve the rdmo options from MaRDMOConfig."""
    return apps.get_app_config("MaRDMO").items

def get_properties():
    """Retrieve the rdmo options from MaRDMOConfig."""
    return apps.get_app_config("MaRDMO").properties

def get_questions(question_set):
    """Retrieve the questions dictionary from MaRDMOConfig."""
    return apps.get_app_config("MaRDMO").questions[question_set]

def get_general_item_url():
    """Get general Wikibase Item URL from Wikibase URL"""
    return f"{endpoint['mardi']['uri']}/wiki/Item:"

def get_data(file_name):
    """Get Data from JSON File"""
    path = os.path.join(os.path.dirname(__file__), file_name)
    with open(path, "r", encoding="utf-8") as json_file:
        data = json.load(json_file)
    return data

def get_sparql_query(file_name):
    """Get Data from SPARQL File"""
    path = os.path.join(os.path.dirname(__file__), file_name)
    with open(path, "r", encoding="utf-8") as sparql_file:
        query = sparql_file.read()
    return query

def get_id(project, uri, keys):
    """Get Set of User requested Identifiers for specific URI"""
    values = project.values.filter(snapshot=None, attribute=Attribute.objects.get(uri=uri))
    identifiers = []
    if len(keys) == 1:
        for value in values:
            identifier = getattr(value, keys[0])
            if isinstance(identifier, str) and '|' in identifier:
                identifier = identifier.split('|')[0]
            identifiers.append(identifier)
    else:
        for value in values:
            identifier = []
            for key in keys:
                identifier.append(getattr(value, key))
            identifiers.append(identifier)
    return identifiers

def get_answers(project, val, config):

    """Unified function to get user answers into dictionary."""

    val.setdefault(config["key1"], {})

    try:
        values = project.values.filter(
            snapshot=None,
            attribute=Attribute.objects.get(uri = f"{BASE_URI}{config['uri']}")
            )
    except Attribute.DoesNotExist:
        values = []

    if not (config["key1"] or config["key2"]):
        return val

    for value in values:

        # Set Prefix IDX
        prefix_idx = None
        if value.set_prefix:
            prefix_idx = int(value.set_prefix.split('|')[0])

        # Set Flags
        flags = (
                 bool(config["set_prefix"]),
                 bool(config["set_index"]),
                 bool(config["collection_index"]),
                 bool(config["external_id"]),
                 bool(config["option_text"]),
                )

        # Set Attribute
        attribute = 'option_uri' if value.option else 'text' if value.text else None

        if not attribute:
            # Ignore if not Attribute Set
            continue

        # Get Flag Combo Handler
        handler = flag_dict[flags]

        # Get Entry and Path
        entry, path = handler(value, attribute, config, prefix_idx)

        # Generate nested Dict Entry
        nested_set(data=val,
                   path=path,
                   entry=entry)

    return val



'''MaRDMO Application File'''

from django.apps import AppConfig

class MaRDMOConfig(AppConfig):
    '''MaRDMO Configuration'''

    name = 'MaRDMO'
    label = 'MaRDMO'
    verbose_name = 'MaRDMO Plugin'

    def __init__(self, app_name, app_module):
        super().__init__(app_name, app_module)
        self.questions = None
        self.mathmoddb = None
        self.mathalgodb = None
        self.options = None
        self.items = None
        self.properties = None

    def ready(self):

        from .getters import get_data

        self.questions = {
            'algorithm': get_data('algorithm/data/questions.json'),
            'model': get_data('model/data/questions.json'),
            'publication': get_data('publication/data/questions.json'),
            'workflow': get_data('workflow/data/questions.json'),
            'search': get_data('search/data/questions.json'),
        }
        self.mathmoddb = get_data('model/data/mapping.json')
        self.mathalgodb = get_data('algorithm/data/mapping.json')
        self.options = get_data('data/options.json')
        self.items = get_data('data/items.json')
        self.properties = get_data('data/properties.json')

        from . import router



'''Module containing Handlers for the Algorithm Documentation'''

from .constants import PROPS
from .models import Benchmark, Software, Problem, Algorithm

from ..config import BASE_URI, endpoint
from ..getters import get_mathalgodb, get_questions, get_sparql_query
from ..queries import query_sparql
from ..adders import (
    add_basics,
    add_entities,
    add_relations_static,
    add_relations_flexible,
    add_references
)

class Information:
    '''Class containing functions, querying external sources for specific
       entities and integrating the related metadata into the questionnaire.'''

    def __init__(self):
        # Load shared data once
        self.questions = get_questions("algorithm") | get_questions('publication')
        self.mathalgodb = get_mathalgodb()
        self.base = BASE_URI

    def benchmark(self, instance):
        '''Benchmark Information'''

        # Benchmark specific Questions.
        benchmark = self.questions["Benchmark"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = "Benchmark",
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, identifier = instance.external_id.split(':')

        # If Item from MathModDB, query relations and load MathModDB Vocabulary
        query = get_sparql_query('algorithm/queries/benchmark.sparql').format(identifier)
        results = query_sparql(query, endpoint[source]['sparql'])

        # Stop if no Results retrieved from external source
        if not results:
            return

        # Get Benchmark Data from Query
        data = Benchmark.from_query(results)

        # Add References to Questionnaire
        add_references(
            project = instance.project,
            data = data,
            uri = f'{self.base}{benchmark["Reference"]["uri"]}',
            set_prefix = instance.set_index
        )

        # Add Publications to Questionnaire
        add_entities(
            project = instance.project,
            question_set = f'{self.base}{self.questions["Publication"]["uri"]}',
            datas = data.publications,
            source = source,
            prefix = 'P'
        )

    def software(self, instance):
        '''Software Information'''

        # Software specific Questions.
        software = self.questions["Software"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = "Software",
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, identifier = instance.external_id.split(':')

        query = get_sparql_query('algorithm/queries/software.sparql').format(identifier)
        results = query_sparql(query, endpoint[source]['sparql'])

        # Stop if no Results retrieved from external source
        if not results:
            return

        # Structure Results and load MathAlgoDB
        data = Software.from_query(results)

        # Add References to Questionnaire
        add_references(
            project = instance.project,
            data = data,
            uri = f'{self.base}{software["Reference"]["uri"]}',
            set_prefix = instance.set_index
        )

        # Add Benchmarks to Questionnaire
        add_relations_static(
            project = instance.project,
            data = data,
            props = {
                'keys': PROPS['S2B']
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relatant': f'{self.base}{software["BRelatant"]["uri"]}'
            }
        )

        # Add Publications to Questionnaire
        add_entities(
            project = instance.project,
            question_set = f'{self.base}{self.questions["Publication"]["uri"]}',
            datas = data.publications,
            source = source,
            prefix = 'P'
        )

    def problem(self, instance):
        '''Algorithmic Problem Information'''

        # Algorithmic Problem specific Questions.
        problem = self.questions["Problem"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = "Problem",
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, identifier = instance.external_id.split(':')

        query = get_sparql_query('algorithm/queries/problem.sparql').format(identifier)
        results = query_sparql(query, endpoint[source]['sparql'])

        # Stop if no Results retrieved from external source
        if not results:
            return

        # Structure Data and load MathAlgoDB
        data = Problem.from_query(results)

        # Add Benchmarks to Questionnaire
        add_relations_static(
            project = instance.project,
            data = data,
            props = {
                'keys': PROPS['P2B']
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relatant': f'{self.base}{problem["BRelatant"]["uri"]}'
            }
        )

        # Add Relations between Algorithmic Problems to Questionnaire
        add_relations_flexible(
            project = instance.project,
            data = data,
            props = {
                'keys': PROPS['Problem'],
                'mapping': self.mathalgodb,
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relation': f'{self.base}{problem["IntraClassRelation"]["uri"]}',
                'relatant': f'{self.base}{problem["IntraClassElement"]["uri"]}',
            },
        )

    def algorithm(self, instance):
        '''Algorithm Information'''

        # Algorithm specific Questions.
        algorithm = self.questions["Algorithm"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = "Algorithm",
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, identifier = instance.external_id.split(':')

        query = get_sparql_query('algorithm/queries/algorithm.sparql').format(identifier)
        results = query_sparql(query, endpoint[source]['sparql'])

        # Stop if no Results retrieved from external source
        if not results:
            return

        # Structure Data and load MathAlgoDB
        data = Algorithm.from_query(results)

        # Add Algorithmic Problems to Questionnaire
        add_relations_static(
            project = instance.project,
            data = data,
            props = {
                'keys': PROPS['A2P']
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relatant': f'{self.base}{algorithm["PRelatant"]["uri"]}'
            }
        )

        # Add Softwares to Questionnaire
        add_relations_static(
            project = instance.project,
            data = data,
            props = {
                'keys': PROPS['A2S']
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relatant': f'{self.base}{algorithm["SRelatant"]["uri"]}'
            }
        )

        # Add Relations between Algorithms to Questionnaire
        add_relations_flexible(
            project = instance.project,
            data = data,
            props = {
                'keys': PROPS['Algorithm'],
                'mapping': self.mathalgodb,
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relation': f'{self.base}{algorithm["IntraClassRelation"]["uri"]}',
                'relatant': f'{self.base}{algorithm["IntraClassElement"]["uri"]}',
            },
        )

        # Add Publications to Questionnaire
        add_entities(
            project = instance.project,
            question_set = f'{self.base}{self.questions["Publication"]["uri"]}',
            datas = data.publications,
            source = source,
            prefix = 'P'
        )



'''Worker Module to collect Publication Metadata'''

from rdmo.projects.models import Value
from rdmo.domain.models import Attribute

from .constants import PUBLICATIONS, JOURNALS, AUTHORS, LANGUAGES
from .utils import generate_label, get_citation
from .sparql import queryPublication
from .models import Publication

from ..config import BASE_URI, endpoint
from ..getters import get_questions, get_properties
from ..queries import query_sparql
from ..helpers import value_editor

class PublicationRetriever:
    '''Retrieve Metadata from MaRDI Portal, Wikidata, MathAlgoDB and other
       sources like Crossref, DataCite, zbMath, DOI, and ORCid for Workflow,
       Model, and Algorithm documentation.'''

    # Get Publication-related Questions
    questions = get_questions('publication')

    def workflow_or_model(self, project, snapshot, answers, options):
        '''Function retrieving Publication Information for workflow and model documentation'''

        for key in answers.get('publication', {}):

            if str(project.catalog).endswith('mardmo-interdisciplinary-workflow-catalog'):
                # Ignore references for individual triple in workflow catalog
                if answers['publication'][key]['workflow'] != options['Yes']:
                    continue

            # If ID is missing (not answered or deleted)
            if not answers['publication'][key].get('ID'):
                continue

            #If User selected Publication from Wikidata...
            if answers['publication'][key]['ID'].startswith('wikidata'):
                #...use the DOI...
                doi = answers['publication'][key].get('reference', {}).get(0, ['',''])[1]
                #...to query MaRDI Portal.
                query = queryPublication['MaRDI']['DOI'].format(doi, **get_properties())
                results = query_sparql(query, endpoint['mardi']['sparql'])
                if results:
                    #If Publication found in MaRDI Portal...
                    data = Publication.from_query(results)
                    #...add data to Questionnaire and...
                    value_editor(
                        project = project,
                        uri = f'{BASE_URI}{self.questions["Publication"]["ID"]["uri"]}',
                        info = {
                            'text': f"{data.label} ({data.description}) [mardi]" , 
                            'external_id': data.id, 
                            'set_index': key
                        }
                    )
                    #...ouput dictionary.
                    answers['publication'][key]['ID'] = data.id
                    answers['publication'][key]['Name'] = data.label
                    answers['publication'][key]['Description'] = data.description

            #If User selected a Publication from MathAlgoDB or did not find it...
            elif answers['publication'][key]['ID'].startswith(('mathalgodb', 'not found')):
                #...but provided a DOI.
                if answers['publication'][key].get('reference', {}).get(0, ['',''])[1]:
                    #Clean potential old data...
                    for question in PUBLICATIONS | LANGUAGES | JOURNALS | AUTHORS:
                        Value.objects.filter(
                            project = project,
                            snapshot = snapshot,
                            attribute_id = Attribute.objects.get(
                                uri = f'{BASE_URI}{self.questions["Publication"][question]["uri"]}'
                            ),
                            set_index = key
                        ).delete()
                    #Get the Citation of several ressource.
                    data_all = get_citation(answers['publication'][key]['reference'][0][1].upper())

                    #If Publication available at MaRDI Portal or Wikidata...
                    if data_all.get('mardi') or data_all.get('wikidata'):
                        data = data_all['mardi'] or data_all['wikidata']
                        #...and add data to Questionnaire and...
                        value_editor(
                            project = project,
                            uri = f'{BASE_URI}{self.questions["Publication"]["ID"]["uri"]}',
                            info = {
                                'text':
                                    f"{data.label} ({data.description}) [{data.id.split(':')[0]}]",
                                'external_id': data.id, 
                                'set_index': key
                            }
                        )
                        #...ouput dictionary.
                        answers['publication'][key]['ID'] = data.id
                        answers['publication'][key]['Name'] = data.label
                        answers['publication'][key]['Description'] = data.description

                    #If Publication available at Crossref, Datacite, zbMath or DOI...
                    elif any(data_all.get(k) for k in ("crossref", "datacite", "zbmath", "doi")):
                        data = (
                            data_all.get("crossref")
                            or data_all.get("datacite")
                            or data_all.get("zbmath")
                            or data_all.get("doi")
                        )
                        #...add data to Questionnaire and...
                        for uri, data_key in PUBLICATIONS.items():
                            value_editor(
                                project = project,
                                uri = f'{BASE_URI}{self.questions["Publication"][uri]["uri"]}',
                                info = {
                                    'text': getattr(data, data_key), 
                                    'set_index': key
                                }
                            )

                        for idx, language in enumerate(data.language):
                            for uri, data_key in LANGUAGES.items():
                                value_editor(
                                    project = project,
                                    uri = f'{BASE_URI}{self.questions["Publication"][uri]["uri"]}',
                                    info = {
                                        'text': getattr(language, data_key), 
                                        'collection_index': idx, 
                                        'set_index': key
                                    }
                                )

                        for idx, journal in enumerate(data.journal):
                            for uri, data_key in JOURNALS.items():
                                value_editor(
                                    project = project,
                                    uri = f'{BASE_URI}{self.questions["Publication"][uri]["uri"]}',
                                    info = {
                                        'text': getattr(journal, data_key), 
                                        'collection_index': idx, 
                                        'set_index': key
                                    }
                                )

                        for idx, author in enumerate(data.authors):
                            for uri, data_key in AUTHORS.items():
                                value_editor(
                                    project = project,
                                    uri = f'{BASE_URI}{self.questions["Publication"][uri]["uri"]}',
                                    info = {
                                        'text': getattr(author, data_key), 
                                        'collection_index': idx, 
                                        'set_index': key
                                    }
                                )

                        #...output dictionary.
                        answers['publication'][key]['Name'] = data.title
                        answers['publication'][key]['Description'] = data.description

        return answers

    def algorithm(self, project, answers):
        '''Function retrieving Publication Information for algorithm documentation'''

        # Go through all Publications
        for key in answers['publication']:
            # If ID is missing (not answered or deleted)
            if not answers['publication'][key].get('ID'):
                continue
            # If User selected Publication from MaRDI Portal or Wikidata...
            if answers['publication'][key]['ID'].startswith(('mardi', 'wikidata')):
                #...check if ressource returned a DOI and...
                if answers['publication'][key].get('reference', {}).get(0, ['',''])[1]:
                    #... use the DOI to query MathAlgoDB.
                    query = queryPublication['PublicationMathAlgoDBDOI'].format(
                        answers['publication'][key]['reference'][0][1]
                    )
                    results = query_sparql(query, endpoint['mathalgodb']['sparql'])
                    if results:
                        # If Publication found on MathAlgoDB...
                        data = Publication.from_query(results)
                        #...add data to Questionnaire and...
                        value_editor(
                            project = project,
                            uri = f'{BASE_URI}{self.questions["Publication"]["ID"]["uri"]}',
                            info = {
                                'text': f"{data.label} ({data.description}) [mathalgodb]" , 
                                'external_id': data.id, 
                                'set_index': key
                            }
                        )
                        #...ouput dictionary.
                        answers['publication'][key]['ID'] = data.id
                        answers['publication'][key]['Name'] = data.label
                        answers['publication'][key]['Description'] = data.description

            #If User did not find its publication...
            elif answers['publication'][key]['ID'].startswith('not found'):
                #...but provided a DOI.
                if answers['publication'][key]['reference'].get(0, ['',''])[1]:
                    #Get the Citation of several ressource.
                    data_all = get_citation(answers['publication'][key]['reference'][0][1])
                    #If Publication available at MathAlgoDB...
                    if data_all['mathalgodb']:
                        #...add data to Questionnaire and...
                        value_editor(
                            project = project,
                            uri = f'{BASE_URI}{self.questions["Publication"]["ID"]["uri"]}',
                            info = {
                                'text': 
                                    f"{data_all['mathalgodb'].label} ({data_all['mathalgodb'].description}) [mathalgodb]",
                                'external_id': data_all['mathalgodb'].id,
                                'set_index': key
                            }
                        )
                        #...ouput dictionary or...
                        answers['publication'][key]['ID'] = data_all['mathalgodb'].id
                        answers['publication'][key]['Name'] = data_all['mathalgodb'].label
                        answers['publication'][key]['Description'] = data_all['mathalgodb'].description
                    #if Publication available at MaRDI/Wikidata/Crossref/DataCite/zbMath/DOI.
                    elif any(data_all.get(k) for k in (
                        "mardi", "wikidata", "crossref",
                        "datacite", "zbmath", "doi")
                    ):
                        data = (
                            data_all.get("mardi")
                            or data_all.get("wikidata")
                            or data_all.get("crossref")
                            or data_all.get("datacite")
                            or data_all.get("zbmath")
                            or data_all.get("doi")
                        )
                        #...add data to Questionnaire and...
                        value_editor(
                            project = project,
                            uri =
                                f'{BASE_URI}{self.questions["Publication"]["Name"]["uri"]}',
                            info = {
                                'text': generate_label(data), 
                                'set_index': key
                            }
                        )
                        value_editor(
                            project = project,
                            uri = 
                                f'{BASE_URI}{self.questions["Publication"]["Description"]["uri"]}',
                            info = {
                                'text': data.description, 
                                'set_index': key
                            }
                        )
                        #...output dictionary.
                        answers['publication'][key]['Name'] = generate_label(data)
                        answers['publication'][key]['Description'] = data.description

        return answers



#SPARQL Query Base and Components for Portal Search

query_base="""
SELECT DISTINCT ?label ?qid
WHERE {{
?workflow wdt:{instance of} wd:{research workflow};
          rdfs:label ?label;
          {0}
          {1}
          {2}
{3}
BIND(STRAFTER(STR(?workflow),STR(wd:)) AS ?qid).
}}
LIMIT 10"""

quote_sparql = "schema:description ?quote;"
res_obj_sparql = "FILTER(CONTAINS(lcase(str(?quote)), '{}'@en)).\n"
res_disc_sparql = "wdt:{field of work} wd:{0};\n"
mmsio_sparql = "wdt:{uses} wd:{0};\n"

query_base_model="""
SELECT DISTINCT ?label ?qid
WHERE {{
?model wdt:{instance of} wd:{mathematical model};
          rdfs:label ?label.
{0}
{1}
{2}
{3}
{4}
{5}
BIND(STRAFTER(STR(?model), STR(wd:)) AS ?qid)
}}
LIMIT 10"""

problem_sparql = """
?problem wdt:{modelled by} ?model;
         wdt:{instance of} wd:{research problem};
         rdfs:label ?problemlabel.
FILTER(lang(?problemlabel) = 'en')"""
problem_filter_sparql = """
FILTER(CONTAINS(lcase(str(?problemlabel)), '{0}'))."""
field_sparql = """
?problem wdt:{modelled by} ?model.
wd:{} wdt:{contains} ?problem;
      wdt:{instance of} wd:{academic discipline}.
"""
formulation_sparql = """
?model wdt:{contains} wd:{0}.
wd:{0} wdt:{instance of} wd:{mathematical expression}.
"""
task_sparql = """
?model wdt:{used by} wd:{0}.
wd:{0} wdt:{instance of} wd:{computational task}.
"""
quantity_sparql = """
{{
  ?model wdt:{contains} ?intermediate .
  ?intermediate wdt:{instance of} wd:{mathematical expression}.
}}
UNION
{{
  ?model wdt:{used by} ?intermediate .
  ?intermediate wdt:{instance of} wd:{computational task}.
}}
VALUES ?type {{ wd:{quantity} wd:{kind of quantity} }} .
"""
quantity_filter_sparql = """
{{
  ?intermediate wdt:{contains} wd:{0}.
  wd:{0} wdt:{instance of} ?type .
}}
""" 

query_base_algorithm="""
PREFIX mathalgodb: <https://mardi4nfdi.de/mathalgodb/0.1#>
PREFIX algorithm: <https://mardi4nfdi.de/mathalgodb/0.1/algorithm#>
PREFIX software: <https://mardi4nfdi.de/mathalgodb/0.1/software#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT DISTINCT ?label ?qid
WHERE {{
?algorithm a mathalgodb:algorithm;
          rdfs:label ?label.
{0}
{1}
{2}
BIND(STRAFTER(STR(?algorithm),"#") AS ?qid).
}}
LIMIT 10"""

algorithmic_problem_sparql = """
?algorithm mathalgodb:solves ?problem.
?problem rdfs:label ?problemlabel"""
algorithmic_problem_filter_sparql = """
FILTER(CONTAINS(lcase(str(?problemlabel)), '{}'))."""
software_sparql = """
?algorithm mathalgodb:implementedBy {}.
"""